@@ -1,10 +1,12 @@
 """pypyr context class. Dictionary ahoy."""
 from collections import namedtuple
 from collections.abc import Mapping, Set
+
 from pypyr.dsl import SpecialTagDirective
 from pypyr.errors import KeyInContextHasNoValueError, KeyNotInContextError
 from pypyr.formatting import RecursiveFormatter
-from pypyr.utils import asserts, expressions, types
+from pypyr.moduleloader import _ChainMapPretendDict
+from pypyr.utils import asserts, types
 
 ContextItemInfo = namedtuple('ContextItemInfo',
                              ['key',
@@ -27,7 +29,6 @@ class Context(dict):
         pipeline_name (str): name of pipeline that is currently running
         working_dir (path-like): working directory path. Either CWD or
                                  initialized from the cli --dir arg.
-
     """
 
     # I *think* instantiating formatter at class level is fine - far as I can
@@ -36,6 +37,35 @@ class Context(dict):
     # https://github.com/python/cpython/blob/master/Lib/string.py
     formatter = RecursiveFormatter(special_types=SpecialTagDirective)
 
+    def __init__(self, *args, **kwargs):
+        """Initialize context."""
+        super().__init__(*args, **kwargs)
+        # __builtins__ are in _ChainMapPretendDict, not in _pystring_globals
+        self._pystring_globals = {}
+        # working on assumption context more frequent lookup than builtins.
+        # here context can go 1st, because eval expressions can't update the
+        # namespace like exec does.
+        self._pystring_namespace = _ChainMapPretendDict(self,
+                                                        self._pystring_globals)
+
+    def __getstate__(self):
+        """Remove namespace from pickle serialization."""
+        state = self.__dict__.copy()
+        # no need to persist builtins - will rehydrate these on setstate.
+        # do want to keep any custom py imports, though.
+        del state['_pystring_namespace']
+        return state
+
+    def __setstate__(self, state):
+        """Rehydrate from pickle will fail on ChainMap coz invocation order.
+
+        Thus custom override to set pystring globals 1st, then namespace with
+        self ref.
+        """
+        self.__dict__.update(state)
+        self._pystring_namespace = _ChainMapPretendDict(self,
+                                                        self._pystring_globals)
+
     def __missing__(self, key):
         """Throw KeyNotInContextError rather than KeyError.
 
@@ -200,7 +230,7 @@ class Context(dict):
             self.assert_key_type_value(context_item, caller, extra_error_text)
 
     def get_eval_string(self, input_string):
-        """Dynamically evaluates the input_string expression.
+        """Dynamically evaluates the input_string python expression.
 
         This provides dynamic python eval of an input expression. The return is
         whatever the result of the expression is.
@@ -208,20 +238,25 @@ class Context(dict):
         Use with caution: since input_string executes any arbitrary code object
         the potential for damage is great.
 
-        The eval uses the current context object as the namespace. This means
-        if you have context['mykey'], in the input_string expression you can
-        use the key directly as a variable like this: "mykey == 'mykeyvalue'".
+        The eval unpacks the current context object into the namespace. This
+        means if you have context['mykey'], in the input_string expression you
+        can use the key directly as a variable like this:
+        "mykey == 'mykeyvalue'".
 
         Both __builtins__ and context are available to the eval expression.
 
-        Args:
-            input_string: expression to evaluate.
+        Args: input_string: expression to evaluate.
 
-        Returns:
-            Whatever object results from the string expression valuation.
+        Returns: Whatever object results from the string expression valuation.
 
         """
-        return expressions.eval_string(input_string, self)
+        if input_string:
+            return eval(input_string, self._pystring_namespace)
+        else:
+            # Empty input raises cryptic EOF syntax err, this more human
+            # friendly
+            raise ValueError('input expression is empty. It must be a valid '
+                             'python expression instead.')
 
     def get_formatted(self, key):
         """Return formatted value for context[key].
@@ -557,6 +592,25 @@ class Context(dict):
         # first iteration starts at context dict root
         merge_recurse(self, add_me)
 
+    def pystring_globals_clear(self):
+        """Clear the pystring globals namespace."""
+        self._pystring_globals.clear()
+
+    def pystring_globals_update(self, *args, **kwargs):
+        """Update the pystring globals namespace with values from other.
+
+        Args:
+            *args/**kwargs:
+                - iterable of key/value pairs
+                - dict
+
+        Returns:
+            Length of updated pystring globals.
+        """
+        # pystring_globals initialized to {} on Context init, no None worries.
+        self._pystring_globals.update(*args, **kwargs)
+        return len(self._pystring_globals)
+
     def set_defaults(self, defaults):
         """Set defaults in context if keys do not exist already.
 
