@@ -11,12 +11,12 @@ from pypyr.errors import (Call,
                           PipelineDefinitionError,
                           Stop)
 from pypyr.cache.stepcache import step_cache
-from pypyr.utils import expressions, poll
+from pypyr.utils import poll
 
 # use pypyr logger to ensure loglevel is set correctly
 logger = logging.getLogger(__name__)
 
-# ------------------------ custom yaml tags -----------------------------------
+# region custom yaml tags
 
 
 class SpecialTagDirective:
@@ -199,7 +199,7 @@ class PyString(SpecialTagDirective):
     def get_value(self, context):
         """Run python eval on the input string."""
         if self.value:
-            return expressions.eval_string(self.value, context)
+            return context.get_eval_string(self.value)
         else:
             # Empty input raises cryptic EOF syntax err, this more human
             # friendly
@@ -230,7 +230,7 @@ class SicString(SpecialTagDirective):
         """Simply return the string as is, the whole point of a sic string."""
         return self.value
 
-# ------------------------ END custom yaml tags -------------------------------
+# endregion custom yaml tags
 
 
 class Step:
@@ -435,12 +435,11 @@ class Step:
         # execution.
         foreach = context.get_formatted_value(self.foreach_items)
 
-        foreach_length = len(foreach)
-
-        logger.info("foreach decorator will loop %s times.", foreach_length)
+        iteration_count = 0
 
         for i in foreach:
             logger.info("foreach: running step %s", i)
+            iteration_count = iteration_count + 1
             # the iterator must be available to the step when it executes
             context['i'] = i
             self.for_counter = i
@@ -450,7 +449,7 @@ class Step:
             self.run_conditional_decorators(context)
             logger.debug("foreach: done step %s", i)
 
-        logger.debug("foreach decorator looped %s times.", foreach_length)
+        logger.info("foreach decorator looped %s times.", iteration_count)
         logger.debug("done")
 
     def invoke_step(self, context):
@@ -780,7 +779,7 @@ class RetryDecorator:
 
         logger.debug("done")
 
-    def exec_iteration(self, counter, context, step_method):
+    def exec_iteration(self, counter, context, step_method, max):
         """Run a single retry iteration.
 
         This method abides by the signature invoked by poll.while_until_true,
@@ -796,6 +795,7 @@ class RetryDecorator:
             step_method: (method/function) This is the method/function that
                          will execute on every loop iteration. Signature is:
                          function(context)
+            max: int. Execute step_method function up to this limit
 
          Returns:
             bool. True if step execution completed without error.
@@ -815,8 +815,8 @@ class RetryDecorator:
             # else, not errors per se.
             raise
         except Exception as ex_info:
-            if self.max:
-                if counter == self.max:
+            if max:
+                if counter == max:
                     logger.debug("retry: max %s retries exhausted. "
                                  "raising error.", counter)
                     # arguably shouldn't be using errs for control of flow.
@@ -892,7 +892,8 @@ class RetryDecorator:
         if poll.while_until_true(interval=sleep,
                                  max_attempts=max)(
                 self.exec_iteration)(context=context,
-                                     step_method=step_method
+                                     step_method=step_method,
+                                     max=max
                                      ):  # pragma: no cover
             logger.debug("retry loop complete, reporting success.")
 
